def liczba_dni_w_roku(rok):
    if rok % 4 == 0:
        dni_rok = 366
    else:
        dni_rok = 365

    return dni_rok


def diff_month(d1, d2):
    return (d1.year - d2.year) * 12 + d1.month - d2.month



import datetime as dt
from dataclasses import dataclass

from dateutil.relativedelta import relativedelta
from scipy import optimize

from procenty.utils import liczba_dni_w_roku

"""
Klasa Lokata:
    * kwota: kwota lokaty
    * oprocentowanie: oprocentowanie lokaty
    * czas: czas lokaty
    * kapitalizacja: kapitalizacja lokaty
"""


@dataclass
class Lokata:
    kwota: float  # Kwota lokaty
    oprocentowanie: float  # Oprocentowanie lokaty (np. 0.05 dla 5%)
    czas: int  # Czas lokaty w miesiącach
    kapitalizacja: int  # Kapitalizacja lokaty (np. 1 - roczna, 12 - miesięczna)

    def __post_init__(self):

        # Oblicz liczbę okresów kapitalizacji
        liczba_okresow = self.czas / (12 / self.kapitalizacja)
        # Oblicz stopę procentową na okres kapitalizacji
        stopa_okresowa = self.oprocentowanie / self.kapitalizacja
        # Oblicz przyszłą wartość lokaty
        self._przyszla_wartosc = self.kwota * (1 + stopa_okresowa) ** liczba_okresow

    def oblicz_zysk(self):
        return self._przyszla_wartosc - self.kwota

    def przyszla_wartosc(self):
        return self._przyszla_wartosc


def npv(rate, cash_flows):
    return sum([cf / (1 + rate) ** i for i, cf in enumerate(cash_flows)])


def irr(cash_flows):
    rate = 0.1
    while npv(rate, cash_flows) > 0:
        rate += 0.01
    return rate


def secant_method(tol, f, x0):
    """
    Solve for x where f(x)=0, given starting x0 and tolerance.

    Arguments
    ----------
    tol: tolerance as percentage of final result. If two subsequent x values are with tol percent, the function will return.
    f: a function of a single variable
    x0: a starting value of x to begin the solver

    Notes
    ------
    The secant method for finding the zero value of a function uses the following formula to find subsequent values of x.

    x(n+1) = x(n) - f(x(n))*(x(n)-x(n-1))/(f(x(n))-f(x(n-1)))

    Warning
    --------
    This implementation is simple and does not handle cases where there is no solution. Users requiring a more robust version should use scipy package optimize.newton.

    """

    x1 = x0 * 1.1
    while abs(x1 - x0) / abs(x1) > tol:
        x0, x1 = x1, x1 - f(x1) * (x1 - x0) / (f(x1) - f(x0))
    return x1


def xnpv(rate, cashflows):
    """
    Calculate the net present value of a series of cashflows at irregular intervals.

    Arguments
    ---------
    * rate: the discount rate to be applied to the cash flows
    * cashflows: a list object in which each element is a tuple of the form (date, amount), where date is a python datetime.date object and amount is an integer or floating point number. Cash outflows (investments) are represented with negative amounts, and cash inflows (returns) are positive amounts.

    Returns
    -------
    * returns a single value which is the NPV of the given cash flows.

    Notes
    ---------------
    * The Net Present Value is the sum of each of cash flows discounted back to the date of the first cash flow. The discounted value of a given cash flow is A/(1+r)**(t-t0), where A is the amount, r is the discout rate, and (t-t0) is the time in years from the date of the first cash flow in the series (t0) to the date of the cash flow being added to the sum (t).
    * This function is equivalent to the Microsoft Excel function of the same name.

    """

    chron_order = sorted(cashflows, key=lambda x: x[0])
    t0 = chron_order[0][0]  # t0 is the date of the first cash flow

    return sum([cf / (1 + rate) ** ((t - t0).days / 365.0) for (t, cf) in chron_order])


def xirr(cashflows, guess=0.1):
    """
    Calculate the Internal Rate of Return of a series of cashflows at irregular intervals.

    Arguments
    ---------
    * cashflows: a list object in which each element is a tuple of the form (date, amount), where date is a python datetime.date object and amount is an integer or floating point number. Cash outflows (investments) are represented with negative amounts, and cash inflows (returns) are positive amounts.
    * guess (optional, default = 0.1): a guess at the solution to be used as a starting point for the numerical solution.

    Returns
    --------
    * Returns the IRR as a single value

    Notes
    ----------------
    * The Internal Rate of Return (IRR) is the discount rate at which the Net Present Value (NPV) of a series of cash flows is equal to zero. The NPV of the series of cash flows is determined using the xnpv function in this module. The discount rate at which NPV equals zero is found using the secant method of numerical solution.
    * This function is equivalent to the Microsoft Excel function of the same name.
    * For users that do not have the scipy module installed, there is an alternate version (commented out) that uses the secant_method function defined in the module rather than the scipy.optimize module's numerical solver. Both use the same method of calculation so there should be no difference in performance, but the secant_method function does not fail gracefully in cases where there is no solution, so the scipy.optimize.newton version is preferred.

    """

    # return secant_method(0.0001,lambda r: xnpv(r,cashflows),guess)
    return optimize.newton(lambda r: xnpv(r, cashflows), guess)


@dataclass
class RRSO:
    wyplata: float
    raty: list
    rrso_0: float

    def right_side(self, rrso):

        R = 0
        for i, r in enumerate(self.raty):
            R += float(r["rata"]) / ((1 + rrso) ** ((i + 1) / 12))
        return R

    def oblicz_rrso(self):
        rrso = self.rrso_0
        l_granica = 0
        r_granica = 10

        rs = self.right_side(rrso)

        i = 0

        while abs(self.wyplata - rs) > 0.0001:

            if self.wyplata > rs:
                r_granica = rrso
            else:
                l_granica = rrso

            rrso = (l_granica + r_granica) / 2

            rs = self.right_side(rrso)

            i += 1
            if i > 1000:
                raise Exception("Za dużo iteracji")
                return "N/A"

        return rrso


def mpkk(K, N, data_start):
    # 28.12.2020 - 04.01.2021
    # 29,30,31,1,2,3,4
    # 3 + 4 = 7

    data_koniec = data_start + relativedelta(months=N)

    # get year out of data_start
    rok_start = data_start.year
    rok_koniec = data_koniec.year

    if rok_start == rok_koniec:
        dni_rok = liczba_dni_w_roku(rok_start)
        dni = (data_koniec - data_start).days
        # print(f"ile dni {dni}")
        mpkk = K * 0.1 + K * dni / dni_rok * 0.1
    else:
        wspolczynnik = 0
        for rok in range(rok_start, rok_koniec + 1):
            # sprawdz czy rok startowy jest przestepny
            if rok == rok_start:
                dni_rok = liczba_dni_w_roku(rok)
                dni = (dt.datetime(rok, 12, 31) - data_start).days
                wspolczynnik = dni / dni_rok
            elif rok == rok_koniec:
                dni_rok = liczba_dni_w_roku(rok)
                dni = (data_koniec - dt.datetime(rok - 1, 12, 31)).days
                wspolczynnik += dni / dni_rok
            else:
                wspolczynnik += 1

        mpkk = K * 0.1 + K * wspolczynnik * 0.1

    return mpkk



import copy
import datetime as dt
from dataclasses import dataclass

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta
from scipy.interpolate import interp1d

from procenty.stopy import Krzywa


@dataclass
class Inflacja:
    krzywa: Krzywa
    miesiac_ref: dt.datetime

    def __post_init__(self):

        self.inflacja_miesiace = self.krzywa.podzial_miesiac()

        self.inflator = self._oblicz_inflator()

    def _oblicz_inflator(self):
        min_miesiac = min(self.inflacja_miesiace, key=lambda x: x[0])[0]
        max_miesiac = max(self.inflacja_miesiace, key=lambda x: x[0])[0]

        # print(f"min_miesiac: {min_miesiac}, max_miesiac: {max_miesiac}, ref: {self.miesiac_ref}")

        if self.miesiac_ref < min_miesiac or self.miesiac_ref > max_miesiac:
            raise ValueError(
                f"Miesac ref spoza przedzialu danych. min_miesiac: {min_miesiac}, max_miesiac: {max_miesiac}, ref: {self.miesiac_ref}"
            )

        # miesiace przed miesiacem referencyjnym

        miesiac_i = self.miesiac_ref - relativedelta(months=1)

        inflator = 1
        inflatory = []

        inflatory.append((self.miesiac_ref, 1))

        i = 0
        while miesiac_i >= min_miesiac:
            i += 1

            mi = miesiac_i.strftime("%Y-%m")

            for m in self.inflacja_miesiace:
                m0 = m[0].strftime("%Y-%m")
                # print(f"m0: {m0}, mi: {mi}")

                if mi == m0:
                    inflator = inflator * m[1]
                    inflatory.append((copy.copy(miesiac_i), copy.copy(inflator)))
                    break

            miesiac_i = miesiac_i - relativedelta(months=1)
            # raise Exception(f"Brak danych dla miesiaca: {miesiac_i}")

        # miesiace po miesiacu referencyjnym
        miesiac_i = self.miesiac_ref + relativedelta(months=1)
        inflator = 1

        i = 0
        while miesiac_i <= max_miesiac:
            i += 1
            mi = miesiac_i.strftime("%Y-%m")
            for m in self.inflacja_miesiace:
                m0 = m[0].strftime("%Y-%m")
                if mi == m0:
                    inflator = inflator * (1.0 / m[1])
                    inflatory.append((copy.copy(miesiac_i), copy.copy(inflator)))
                    break

            miesiac_i = miesiac_i + relativedelta(months=1)
            # raise Exception(f"Brak danych dla miesiaca: {miesiac_i}")

        inflatory = sorted(inflatory, key=lambda x: x[0])

        return inflatory

    def urealnij(self, data, wartosc) -> float:

        miesiac = data.strftime("%Y-%m")

        for i in self.inflator:
            if i[0].strftime("%Y-%m") == miesiac:
                # print(wartosc, i[1])
                return float(wartosc) * float(i[1])

        raise ValueError(f"Brak danych dla miesiaca: {miesiac}")


@dataclass
class InflacjaMiesiac:

    data_start: dt.datetime
    okresy: int
    liczba_wakacji: int
    dane: list
    prognoza: list

    def __post_init__(self):

        self.data_koniec = self.data_start + relativedelta(
            months=(self.okresy + self.liczba_wakacji)
        )

        if self.dane:
            self.df = pd.DataFrame(self.dane)

            # self.df.index = pd.to_datetime(self.df.index, format='%Y-%m-%d')
            self.df.set_index("miesiac", inplace=True)
            self.df["wartosc"] = pd.to_numeric(self.df["wartosc"])
            self.df.index = pd.to_datetime(self.df.index, format="%Y-%m")

            self.max_inflacja_real = self.df.index.max()

            if self.data_koniec > self.max_inflacja_real:

                if self.prognoza:
                    self.prognoza = [
                        (dt.datetime.strptime(p[0], "%d/%m/%Y"), p[1])
                        for p in self.prognoza
                        if dt.datetime.strptime(p[0], "%d/%m/%Y") > self.df.index.max()
                    ]

                self.prognoza.append(
                    (self.df.index.max(), self.df.loc[self.df.index.max(), "wartosc"])
                )

                # interpolation
                dates, values = zip(*self.prognoza)

                # Convert dates to numerical values representing time or elapsed time
                timestamps = np.array(
                    [(date - self.max_inflacja_real).total_seconds() for date in dates]
                )

                # Create an interpolation function using scipy.interpolate.interp1d
                self.interpolation_function = interp1d(timestamps, values)

                # Convert the new date to a numerical value
                end_timestamp = (
                    self.data_koniec - self.max_inflacja_real
                ).total_seconds()

                # Use the interpolation function to estimate the value at the new date
                inflacja_value_koniec = self.interpolation_function(
                    end_timestamp
                ).item()

                self.prognoza = [
                    p
                    for p in self.prognoza
                    if p[0] < self.data_koniec and p[0] > self.df.index.max()
                ]

                self.prognoza.append((self.data_koniec, inflacja_value_koniec))

                new_df = pd.DataFrame(self.prognoza, columns=["miesiac", "wartosc"])
                # new_df['Date'] = pd.to_datetime(new_df['Date'])
                new_df.set_index("miesiac", inplace=True)

                # Concatenate the original DataFrame and the new DataFrame
                self.dff = pd.concat([self.df, new_df])
            else:
                self.dff = self.df[self.df.index <= self.data_koniec]

        else:

            if self.prognoza:
                self.prognoza = [
                    (dt.datetime.strptime(p[0], "%d/%m/%Y"), p[1])
                    for p in self.prognoza
                ]

            # interpolation
            dates, values = zip(*self.prognoza)

            min_date = min(dates)

            # print(f"min_date: {min_date}, data start: {self.data_start}, data koniec: {self.data_koniec}")

            # Convert dates to numerical values representing time or elapsed time
            timestamps = np.array([(date - min_date).total_seconds() for date in dates])

            # Create an interpolation function using scipy.interpolate.interp1d
            self.interpolation_function = interp1d(timestamps, values)

            # Convert the new date to a numerical value
            start_timestamp = (self.data_start - min_date).total_seconds()

            inflacja_value_start = self.interpolation_function(start_timestamp).item()

            # Convert the new date to a numerical value
            end_timestamp = (self.data_koniec - min_date).total_seconds()

            # Use the interpolation function to estimate the value at the new date
            inflacja_value_koniec = self.interpolation_function(end_timestamp).item()

            self.prognoza.append((self.data_start, inflacja_value_start))
            self.prognoza.append((self.data_koniec, inflacja_value_koniec))

            self.prognoza = [
                p
                for p in self.prognoza
                if p[0] >= self.data_start and p[0] <= self.data_koniec
            ]

            self.dff = pd.DataFrame(self.prognoza, columns=["miesiac", "wartosc"])
            # new_df['Date'] = pd.to_datetime(new_df['Date'])
            self.dff.set_index("miesiac", inplace=True)

        self.dff = self.dff.sort_index()

        self.json_data = [
            {"date": dt.datetime.strftime(index, "%Y-%m-%d"), "value": round(value, 2)}
            for index, value in self.dff["wartosc"].items()
        ]

    def _getInflacja(self, dzien):
        if self.dane:
            if dzien > self.max_inflacja_real:
                # Generate a new date for interpolation

                # Convert the new date to a numerical value
                new_timestamp = (dzien - self.max_inflacja_real).total_seconds()

                # Use the interpolation function to estimate the value at the new date
                inflacja_value = float(self.interpolation_function(new_timestamp))

            else:
                inflacja_value = self._getInflacjaReal(dzien)

        else:
            # Convert the new date to a numerical value
            new_timestamp = (dzien - self.data_start).total_seconds()

            # Use the interpolation function to estimate the value at the new date
            inflacja_value = float(self.interpolation_function(new_timestamp))

        return inflacja_value

    def _getInflacjaReal(self, dzien):
        inflacja_value = self.dff.loc[
            (self.dff.index.year == dzien.year) & (self.dff.index.month == dzien.month),
            "wartosc",
        ]
        return inflacja_value[0]

    def urealnij(self, dates_values_list) -> list:

        start_date = self.data_start

        end_date = max([dv["dzien"] for dv in dates_values_list])

        inflator = {f"{dt.datetime.strftime(start_date, '%Y-%m')}": 1}
        current_date = start_date

        rolling_mult = 1
        while current_date < end_date:
            current_date = current_date + relativedelta(months=1)
            wartosc = self._getInflacja(current_date) / 100.0
            rolling_mult = rolling_mult * wartosc
            inflator[f"{dt.datetime.strftime(current_date, '%Y-%m')}"] = rolling_mult

        new_values = []
        for dv in dates_values_list:
            miesiac = dt.datetime.strftime(dv["dzien"], "%Y-%m")
            wsk_infl = inflator[miesiac]
            new_values.append(
                {"miesiac": miesiac, "wartosc": float(dv["wartosc"]) / wsk_infl}
            )

        return new_values



"""Oblicz swój kredyt."""

import copy
import datetime as dt
import decimal
from dataclasses import dataclass
from decimal import ROUND_HALF_UP, Decimal
from enum import Enum
from typing import Optional

from dateutil.relativedelta import relativedelta
from loguru import logger

import procenty.inwestycja as inv
from procenty.inflacja import Inflacja
from procenty.miary import LiczbaDni

grosze = decimal.Decimal(".01")


class Rodzaj(Enum):
    SPLATA = 3
    OPROCENTOWANIE = 4
    NADPLATA = 2
    SPLATA_CALKOWITA = 6
    TRANSZA = 1
    WAKACJE = 5


@dataclass
class Zdarzenie:
    data: dt.datetime
    rodzaj: Rodzaj
    wartosc: object

    def __lt__(self, other):

        return str(self.data) + str(self.rodzaj.value) < str(other.data) + str(
            other.rodzaj.value
        )


@dataclass
class Kredyt:
    """Klasa do obsługi kredytów."""

    K: Decimal
    N: int
    r: Decimal
    marza: Decimal
    start: dt.datetime
    rodzajRat: str
    splaty_normalne: bool = True
    operacje: Optional[list[Zdarzenie]] = None

    def __post_init__(self):
        self.Kstart = self.K
        self.Nstart = self.N

        self.p = self.r + self.marza

        self.K = self.K.quantize(grosze, ROUND_HALF_UP)
        self.dzien_odsetki: dt.datetime = self.start
        self.zdarzenia: list[Zdarzenie] = []
        
        if self.operacje:
            self.zdarzenia.extend(self.operacje)

        self.odsetki_naliczone = 0
        self.odsetki_naliczone_marza = 0
        self.I = 0

        self.licznik_rat = 0

        self.transze = []

        self.wynik = []
        self.wynik_nadplaty = []

        self.wszystkie_koszty = []
        self.wszystkie_koszty_kwota = 0

        if self.splaty_normalne:
            dni_splaty = [
                self.start + relativedelta(months=i + 1) for i in range(self.N)
            ]
            for dzien_splaty in dni_splaty:
                self.zdarzenia.append(Zdarzenie(dzien_splaty, Rodzaj.SPLATA, 0))

        N_zdarzenie = [z for z in self.zdarzenia if z.rodzaj == Rodzaj.SPLATA]
        if len(N_zdarzenie) != self.N:
            logger.warning(
                f"Kredyt: {self.K}, splaty normalne {self.splaty_normalne}.Nie zgadza się liczba zdarzeń {len(N_zdarzenie)} z N {self.N}"
            )
            raise Exception("Nie zgadza się liczba zdarzeń SPLATA RATY z argumentem N")

        self.kredyt_wynik = self._symuluj()

        if self.rodzajRat == "malejace_met2":
            print("przeliczam---------------------------")
            odsetki = [float(x["odsetki"]) for x in self.kredyt_wynik["raty"]]
            odsetki_marza = [
                float(x["odsetki_marza"]) for x in self.kredyt_wynik["raty"]
            ]
            odsetki_wibor = [
                float(x["odsetki_wibor"]) for x in self.kredyt_wynik["raty"]
            ]

            nowe_odsetki = self.__przelicz_odsetki_mal2(odsetki)
            nowe_odsetki_marza = self.__przelicz_odsetki_mal2(odsetki_marza)
            nowe_odsetki_wibor = self.__przelicz_odsetki_mal2(odsetki_wibor)

            for i in range(len(self.kredyt_wynik["raty"])):
                self.kredyt_wynik["raty"][i]["odsetki"] = str(
                    Decimal(nowe_odsetki[i]).quantize(grosze, ROUND_HALF_UP)
                )
                self.kredyt_wynik["raty"][i]["odsetki_marza"] = nowe_odsetki_marza[i]
                self.kredyt_wynik["raty"][i]["odsetki_wibor"] = nowe_odsetki_wibor[i]

                nowa_rata = Decimal(
                    nowe_odsetki[i] + float(self.kredyt_wynik["raty"][i]["kapital"])
                )
                self.kredyt_wynik["raty"][i]["rata"] = str(
                    nowa_rata.quantize(grosze, ROUND_HALF_UP)
                )

    def __repr__(self) -> str:
        return "kredyt : {}".format(self.K)

    def _dodaj_koszt(self, dzien_koszt: dt.datetime, kwota: Decimal):
        dotychczasowe_koszty = self.wszystkie_koszty_kwota
        self.wszystkie_koszty_kwota = dotychczasowe_koszty + kwota
        self.wszystkie_koszty.append(
            {
                "dzien": dzien_koszt.strftime("%Y-%m-%d"),
                "kwota": float(copy.copy(self.wszystkie_koszty_kwota)),
            }
        )

    def __przelicz_odsetki_mal2(self, ciag_platnosci: list[float]) -> list[float]:

        N = len(ciag_platnosci)
        sum_O = sum(ciag_platnosci)

        o_N = ciag_platnosci[-1]

        # Wyznaczanie współczynnika A
        # A = (2 * sum_O - N * o_N) / (N * (N + 1))

        a1 = N * ((1 + N) / 2) - N * N
        A = (sum_O - N * o_N) / a1

        # Wyznaczanie współczynnika B
        B = o_N - A * N

        # Tworzenie nowego ciągu odsetek
        nowy_ciag = [A * i + B for i in range(1, N + 1)]

        return nowy_ciag

    def pokaz_koszty(self):
        return self.wszystkie_koszty

    def pokaz_koszty_real(self, inflator: Inflacja):
        wynik = []
        for koszt in self.wszystkie_koszty:
            wynik.append(
                {
                    "dzien": koszt["dzien"],
                    "kwota": inflator.urealnij(
                        dt.datetime.strptime(koszt["dzien"], "%Y-%m-%d"), koszt["kwota"]
                    ),
                }
            )
        return wynik

    def wyswietl(self, dzien_raty):

        grosze = decimal.Decimal(".01")
        return "dzien: {}, K: {} zł, odsetki: {} zł, rata: {} zł".format(
            dzien_raty,
            self.K.quantize(grosze, ROUND_HALF_UP),
            self.odsetki_naliczone.quantize(grosze, ROUND_HALF_UP),
            self.I.quantize(grosze, ROUND_HALF_UP),
        )

    def zapisz_stan(self, dzien_raty):

        data = {
            "dzien": dzien_raty.strftime("%Y-%m-%d"),
            "K": str(self.K.quantize(grosze, ROUND_HALF_UP)),
            "odsetki": str(self.odsetki_naliczone.quantize(grosze, ROUND_HALF_UP)),
            "odsetki_marza": str(
                self.odsetki_naliczone_marza.quantize(grosze, ROUND_HALF_UP)
            ),
            "odsetki_wibor": str(
                self.odsetki_naliczone.quantize(grosze, ROUND_HALF_UP)
                - self.odsetki_naliczone_marza.quantize(grosze, ROUND_HALF_UP)
            ),
            "kapital": str(
                self.I.quantize(grosze, ROUND_HALF_UP)
                - self.odsetki_naliczone.quantize(grosze, ROUND_HALF_UP)
            ),
            "rata": str(self.I.quantize(grosze, ROUND_HALF_UP)),
            "nr_raty": self.licznik_rat,
            "K_po": str(
                (self.K - (self.I - self.odsetki_naliczone)).quantize(
                    grosze, ROUND_HALF_UP
                )
            ),
        }

        self.wynik.append(data)

        return 1

    def oblicz_rate(self) -> Decimal:

        if self.rodzajRat == "rowne":
            k = 12
            do_splaty = self.K
            if self.p > 0:
                L = do_splaty * self.p
                M = k * (1 - pow(k / (k + self.p), self.N))
                I = L / M
            else:
                I = do_splaty / self.N
        elif self.rodzajRat == "malejace":
            I = (self.K / self.N) * (1 + (self.p / 12) * self.N)
        elif self.rodzajRat == "malejace_met2":
            I = (self.K / self.N) * (1 + (self.p / 12) * (self.N + 1) / 2)  # workaround
        else:
            raise Exception("nie ma takich rat")

        return I

    def zmien_oprocentowanie(self, dzien_zmiany: dt.datetime, nowe_r: Decimal):

        o_d = LiczbaDni(
            self.dzien_odsetki.strftime("%Y-%m-%d"),
            dzien_zmiany.strftime("%Y-%m-%d")
        )

        opr = o_d.mnoznik * self.p

        opr_marza = o_d.mnoznik * self.marza

        self.odsetki_naliczone = self.odsetki_naliczone + opr * self.K
        self.odsetki_naliczone_marza = self.odsetki_naliczone_marza + opr_marza * self.K

        self.p = Decimal((nowe_r / 100.0)) + self.marza

        self.dzien_odsetki = dzien_zmiany

    def zrob_nadplate(self, dzien_nadplaty: dt.datetime, kwota: Decimal):

        o_d = LiczbaDni(
            self.dzien_odsetki.strftime("%Y-%m-%d"),
            dzien_nadplaty.strftime("%Y-%m-%d")
        )

        opr = o_d.mnoznik * self.p

        opr_marza = o_d.mnoznik * self.marza

        self.odsetki_naliczone = self.odsetki_naliczone + opr * self.K

        self.odsetki_naliczone_marza = self.odsetki_naliczone_marza + opr_marza * self.K

        if kwota > self.K:
            kwota = self.K
            self.K = Decimal(0)
        else:
            self.K = self.K - kwota

        self.dzien_odsetki = dzien_nadplaty
        grosze = decimal.Decimal(".01")
        self.wynik_nadplaty.append(
            {
                "dzien": str(dzien_nadplaty.strftime("%Y-%m-%d")),
                "kwota": str(kwota.quantize(grosze, ROUND_HALF_UP)),
            }
        )

        self._dodaj_koszt(dzien_nadplaty, kwota)

    def zrob_transze(self, dzien_transzy: dt.datetime, kwota: Decimal):

        o_d = LiczbaDni(
            self.dzien_odsetki.strftime("%Y-%m-%d"),
            dzien_transzy.strftime("%Y-%m-%d")
        )

        opr = o_d.mnoznik * self.p

        opr_marza = o_d.mnoznik * self.marza

        self.odsetki_naliczone = self.odsetki_naliczone + opr * self.K
        self.odsetki_naliczone_marza = self.odsetki_naliczone_marza + opr_marza * self.K

        self.K = self.K + kwota

        self.dzien_odsetki = dzien_transzy

        self.transze.append({"dzien": dzien_transzy, "kwota": kwota})

    def zrob_splate_calkowita(self, dzien_splaty: dt.datetime):

        o_d = LiczbaDni(
            self.dzien_odsetki.strftime("%Y-%m-%d"),
            dzien_splaty.strftime("%Y-%m-%d")
        )

        opr = o_d.mnoznik * self.p

        opr_marza = o_d.mnoznik * self.marza

        self.odsetki_naliczone = (self.odsetki_naliczone + opr * self.K).quantize(
            grosze, ROUND_HALF_UP
        )

        self.odsetki_naliczone_marza = self.odsetki_naliczone_marza + opr_marza * self.K

        kwota = self.K - self.odsetki_naliczone

        self.I = kwota

        self.licznik_rat += 1
        self.zapisz_stan(dzien_splaty)

        self.K = Decimal(0).quantize(grosze, ROUND_HALF_UP)
        self.dzien_odsetki = dzien_splaty
        self._dodaj_koszt(dzien_splaty, kwota)

    def splata_raty(self, dzien_raty: dt.datetime):

        o_d = LiczbaDni(
            self.dzien_odsetki.strftime("%Y-%m-%d"),
            dzien_raty.strftime("%Y-%m-%d")
        )

        opr = o_d.mnoznik * self.p

        opr_marza = o_d.mnoznik * self.marza

        if self.K > 0:
            self.I = self.oblicz_rate().quantize(grosze, ROUND_HALF_UP)
        else:
            self.I = Decimal(0).quantize(grosze, ROUND_HALF_UP)

        self.odsetki_naliczone = (self.odsetki_naliczone + opr * self.K).quantize(
            grosze, ROUND_HALF_UP
        )
        self.odsetki_naliczone_marza = self.odsetki_naliczone_marza + opr_marza * self.K

        # work around
        if self.rodzajRat == "malejace_met2":
            self.I = (self.Kstart / self.Nstart + self.odsetki_naliczone).quantize(
                grosze, ROUND_HALF_UP
            )
            if self.K <= 0:
                self.I = Decimal(0).quantize(grosze, ROUND_HALF_UP)

        if self.odsetki_naliczone > self.I:
            self.I = self.odsetki_naliczone

        # ostatnia rata
        if self.N == 1:
            roznica_na_koniec = self.K - (self.I - self.odsetki_naliczone)
            self.I += roznica_na_koniec

        self.licznik_rat += 1
        self.zapisz_stan(dzien_raty)

        self.K = self.K - (self.I - self.odsetki_naliczone)

        self.odsetki_naliczone = 0
        self.odsetki_naliczone_marza = 0

        self.dzien_odsetki = dzien_raty
        self.N -= 1

        self._dodaj_koszt(dzien_raty, self.I)

    def _symuluj(self):

        for zdarzenie in sorted(self.zdarzenia):

            if zdarzenie.rodzaj == Rodzaj.OPROCENTOWANIE:
                self.zmien_oprocentowanie(zdarzenie.data, zdarzenie.wartosc)
            elif zdarzenie.rodzaj == Rodzaj.SPLATA:
                self.splata_raty(zdarzenie.data)
            elif zdarzenie.rodzaj == Rodzaj.SPLATA_CALKOWITA:
                self.zrob_splate_calkowita(zdarzenie.data)
            elif zdarzenie.rodzaj == Rodzaj.NADPLATA:
                self.zrob_nadplate(zdarzenie.data, zdarzenie.wartosc)
            elif zdarzenie.rodzaj == Rodzaj.TRANSZA:
                self.zrob_transze(zdarzenie.data, zdarzenie.wartosc)

            if self.K == Decimal(0):
                break

        return {
            "raty": self.wynik,
            "nadplaty": self.wynik_nadplaty,
            "kapital_na_koniec": str(self.K.quantize(decimal.Decimal("0.01"))),
        }

    @property
    def raty(self) -> list:
        return self.kredyt_wynik["raty"]

    @property
    def nadplaty(self) -> list:
        return self.kredyt_wynik["nadplaty"]

    @property
    def xirr(self) -> float:
        cashflows = [
            (dt.datetime.strptime(x["dzien"], "%Y-%m-%d"), float(x["rata"]))
            for x in self.kredyt_wynik["raty"]
        ]
        cashflows = cashflows + [
            (dt.datetime.strptime(x["dzien"], "%Y-%m-%d"), float(x["kwota"]))
            for x in self.kredyt_wynik["nadplaty"]
        ]
        cashflows = cashflows + [(self.start, float(-self.Kstart))]
        cashflows = cashflows + [(x["dzien"], float(-x["kwota"])) for x in self.transze]

        return inv.xirr(cashflows)

    @property
    def podsumowanie(self) -> str:
        raty = [float(x["rata"]) for x in self.kredyt_wynik["raty"]]
        ile = len(raty)
        suma = sum(raty)
        nadplaty = [float(x["kwota"]) for x in self.kredyt_wynik["nadplaty"]]
        suma_nadplat = sum(nadplaty)
        odsetki = [float(x["odsetki"]) for x in self.kredyt_wynik["raty"]]
        suma_odsetek = sum(odsetki)

        dane = {
            "raty": self.kredyt_wynik["raty"],
            "info": {
                "suma_rat": str(Decimal(suma).quantize(grosze, ROUND_HALF_UP)),
                "suma_nadplat": str(
                    Decimal(suma_nadplat).quantize(grosze, ROUND_HALF_UP)
                ),
                "suma_odsetek": str(
                    Decimal(suma_odsetek).quantize(grosze, ROUND_HALF_UP)
                ),
                "liczba_rat": ile,
                "xirr": str(self.xirr),
                "K": str(self.Kstart.quantize(grosze, ROUND_HALF_UP)),
            },
        }

        return dane


@dataclass
class KredytPorownanie:
    """Klasa do obsługi kredytów."""

    kredyt: Kredyt
    r: Decimal

    def __post_init__(self):

        self.kredyt = copy.deepcopy(self.kredyt)

        self.kredyt_porownanie = Kredyt(
            self.kredyt.Kstart,
            self.kredyt.Nstart,
            self.r,
            self.kredyt.marza,
            self.kredyt.start,
            self.kredyt.rodzajRat,
            self.kredyt.splaty_normalne,
            self.kredyt.operacje,
        )

    @property
    def xirr(self) -> float:
        return self.kredyt_porownanie.xirr

    @property
    def raty(self) -> list:
        return self.kredyt_porownanie.raty

    @property
    def podsumowanie(self) -> dict:
        return self.kredyt_porownanie.podsumowanie


@dataclass
class KredytSuwak:
    K: Decimal
    N: int
    p: Decimal
    marza: Decimal
    start: dt.datetime
    operacje: Optional[list[Zdarzenie]] = None

    def __post_init__(self):
        self.dzien_odsetki: dt.datetime = self.start
        self.I = 0
        self.odsetki_naliczone = 0

    def zmien_oprocentowanie(self, dzien_zmiany: dt.datetime, nowe_r: Decimal):

        o_d = LiczbaDni(
            self.dzien_odsetki.strftime("%Y-%m-%d"),
            dzien_zmiany.strftime("%Y-%m-%d")
        )

        opr = o_d.mnoznik * self.p

        self.odsetki_naliczone = self.odsetki_naliczone + opr * self.K

        self.p = Decimal((nowe_r * 0.01)) + self.marza

        self.dzien_odsetki = dzien_zmiany

    def splata_raty(self, dzien_raty: dt.datetime):

        o_d = LiczbaDni(
            self.dzien_odsetki.strftime("%Y-%m-%d"),
            dzien_raty.strftime("%Y-%m-%d")
        )

        opr = o_d.mnoznik * self.p

        if self.K > 0:
            self.I = self.oblicz_rate().quantize(grosze, ROUND_HALF_UP)
        else:
            self.I = Decimal(0).quantize(grosze, ROUND_HALF_UP)

        self.odsetki_naliczone = (self.odsetki_naliczone + opr * self.K).quantize(
            grosze, ROUND_HALF_UP
        )

        if self.odsetki_naliczone > self.I:
            self.I = self.odsetki_naliczone

        # ostatnia rata
        if self.N == 1:
            roznica_na_koniec = self.K - (self.I - self.odsetki_naliczone)
            self.I += roznica_na_koniec

        self.K = self.K - (self.I - self.odsetki_naliczone)

        self.odsetki_naliczone = 0

        self.dzien_odsetki = dzien_raty
        self.N -= 1

    def next(self, data: dt.datetime, rata_porownawcza: Decimal):

        # sprawdz czy pomiędzy ostatnią spłatą a datą jest zmiana oprocentowania
        # jeśli tak to zmień oprocentowanie

        if self.N > 0 and self.K > 0:

            if self.operacje:
                for zdarzenie in self.operacje:
                    if zdarzenie.data > self.dzien_odsetki and zdarzenie.data <= data:
                        if zdarzenie.rodzaj == Rodzaj.OPROCENTOWANIE:
                            self.zmien_oprocentowanie(zdarzenie.data, zdarzenie.wartosc)

            self.splata_raty(data)

            nadplata = 0
            if self.I < rata_porownawcza:
                nadplata = rata_porownawcza - self.I
                self.K = self.K - nadplata

        return self.K

    def oblicz_rate(self) -> Decimal:

        k = 12
        do_splaty = self.K
        liczba_rat = self.N
        if self.p > 0:
            L = do_splaty * self.p
            M = k * (1 - pow(k / (k + self.p), liczba_rat))
            I = L / M
        else:
            I = do_splaty / liczba_rat

        return I



import datetime as dt
from decimal import Decimal
from typing import Any

from ..kredyt import Kredyt, Rodzaj, Zdarzenie


def create_kredyt(dane: list[dict[str, Any]], rodzajRat: str):

    r = Decimal(dane["r"] / 100.0)
    marza = Decimal(dane["marza"] / 100.0)
    K = Decimal(dane["K"])
    dni = dane["daty_splaty"]
    N = len(dni)
    start_kredytu = dt.datetime.strptime(dane["start"], "%Y-%m-%d")

    zdarzenia = []

    for dzien_splaty in dane["daty_splaty"]:
        zdarzenia.append(
            Zdarzenie(dt.datetime.strptime(dzien_splaty, "%Y-%m-%d"), Rodzaj.SPLATA, 0)
        )

    if "oprocentowanie" in dane:
        for zmiana_opr in dane["oprocentowanie"]:
            zdarzenia.append(
                Zdarzenie(
                    dt.datetime.strptime(zmiana_opr["dzien"], "%Y-%m-%d"),
                    Rodzaj.OPROCENTOWANIE,
                    zmiana_opr["proc"],
                )
            )

    if "nadplaty" in dane:
        for nadplata in dane["nadplaty"]:
            if nadplata["calkowita"] == True:
                zdarzenia.append(
                    Zdarzenie(
                        dt.datetime.strptime(nadplata["dzien"], "%Y-%m-%d"),
                        Rodzaj.SPLATA_CALKOWITA,
                        Decimal(nadplata["kwota"]),
                    )
                )
            else:
                zdarzenia.append(
                    Zdarzenie(
                        dt.datetime.strptime(nadplata["dzien"], "%Y-%m-%d"),
                        Rodzaj.NADPLATA,
                        Decimal(nadplata["kwota"]),
                    )
                )

    if "transze" in dane:
        for transza in dane["transze"]:
            zdarzenia.append(
                Zdarzenie(
                    dt.datetime.strptime(transza["dzien"], "%Y-%m-%d"),
                    Rodzaj.TRANSZA,
                    Decimal(transza["kapital"]),
                )
            )

    kr = Kredyt(K, N, r, marza, start_kredytu, rodzajRat, False, zdarzenia)

    return kr


def create_kredyt_normalny(dane: list[dict[str, Any]], rodzajRat: str):

    r = Decimal(dane["r"] / 100.0)
    marza = Decimal(dane["marza"] / 100.0)
    K = Decimal(dane["K"])
    N = dane["N"]
    start_kredytu = dt.datetime.strptime(dane["start"], "%Y-%m-%d")

    kr = Kredyt(K, N, r, marza, start_kredytu, rodzajRat, True, [])

    return kr



"""
Moduł z klasami miar finansowych.



"""

from dataclasses import dataclass
from decimal import Decimal
from .utils.inne import liczba_dni_w_roku
import pendulum


@dataclass
class LiczbaDni:
    """Liczy dni do np. wyliczania odsetek.

    Bierze pod uwagę liczbę dni w latach przestępnych.

    Argumenty:
    start -- data w formacie YYYY-MM-DD
    koniec -- data w formacie YYYY-MM-DD

    mnożnik - proporcja dni odsetkowych do liczby dni w roku.
    """

    start: str
    koniec: str


    def __post_init__(self):
        dt_start = pendulum.from_format(self.start, "YYYY-MM-DD")
        dt_koniec = pendulum.from_format(self.koniec, "YYYY-MM-DD")

        if dt_start > dt_koniec:
            raise Exception("Dzien start > Dzien koniec")

        self._dni_odsetkowe = dt_koniec.diff(dt_start).in_days()

        mnoznik = 0

        if dt_start.year == dt_koniec.year:
            mnoznik = self._dni_odsetkowe / liczba_dni_w_roku(dt_start.year)
        else:
            for rok in range(dt_start.year, dt_koniec.year + 1):
                if rok == dt_start.year:
                    mnoznik += pendulum.datetime(dt_start.year + 1, 1, 1).diff(
                        dt_start
                    ).in_days() / liczba_dni_w_roku(dt_start.year)
                elif rok == dt_koniec.year:
                    mnoznik += dt_koniec.diff(
                        pendulum.datetime(dt_koniec.year, 1, 1)
                    ).in_days() / liczba_dni_w_roku(dt_koniec.year)
                else:
                    mnoznik += 1

        self._mnoznik = mnoznik

    @property
    def dni_odsetkowe(self):
        return self._dni_odsetkowe

    @property
    def mnoznik(self) -> Decimal:
        return Decimal(self._mnoznik)


class Zloty:
    """Klasa do obsługi kwot w złotówkach.

    Argumenty:

    kwota -- kwota w złotówkach
    """

    def __init__(self, kwota: float):
        self.kwota = kwota

    def __add__(self, other):
        return Zloty(self.kwota + other.kwota)

    def __sub__(self, other):
        return Zloty(self.kwota - other.kwota)

    def __mul__(self, other):
        return Zloty(self.kwota * other)

    def __truediv__(self, other):
        return Zloty(self.kwota / other)

    def __str__(self):
        return f"{self.kwota} zł"

    def __repr__(self):
        return f"{self.kwota} zł"

    def __eq__(self, other):
        return self.kwota == other.kwota

    def __lt__(self, other):
        return self.kwota < other.kwota

    def __le__(self, other):
        return self.kwota <= other.kwota

    def __gt__(self, other):
        return self.kwota > other.kwota

    def __ge__(self, other):
        return self.kwota >= other.kwota

    def __ne__(self, other):
        return self.kwota != other.kwota



import datetime as dt
import decimal
from dataclasses import dataclass

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta
from pandas.tseries.offsets import BDay
from scipy.interpolate import interp1d
from sqlalchemy import text as sql_text


@dataclass
class WiborInter:
    """interpolowanie wibour dla wartosci, ktorych jeszce nie znamy."""

    db: object
    wibor_typ: str
    data_start: dt.datetime
    okresy: int
    liczba_wakacji: int
    points: list

    def __post_init__(self):
        if self.wibor_typ == "3M":
            self._okres = 3
            wib_rodzaj = "wibor3m"
        elif self.wibor_typ == "6M":
            self._okres = 6
            wib_rodzaj = "wibor6m"

        # self.df = pd.read_csv('obliczeniakredytowe/static/{}'.format(file_name), usecols=[0,1], index_col=0)
        # self.df.index = pd.to_datetime(self.df.index, format='%Y-%m-%d')

        self.df = pd.read_sql(
            sql_text(f"SELECT data, wartosc FROM wibor WHERE rodzaj='{wib_rodzaj}'"),
            con=self.db.engine.connect(),
            index_col="data",
        )

        self.df.index = pd.to_datetime(self.df.index, format="%Y-%m-%d")

        self.df["real"] = "Y"

        self.df.sort_index(inplace=True)

        self.df = self.df[self.df.index >= self.data_start - relativedelta(days=5)]

        self.data_koniec = self.data_start + relativedelta(
            months=(self.okresy + self.liczba_wakacji)
        )

        self.max_wibor_real = self.df.index.max()

        if self.data_koniec > self.max_wibor_real:

            if self.points:
                self.points = [
                    (dt.datetime.strptime(p[0], "%d/%m/%Y"), p[1])
                    for p in self.points
                    if dt.datetime.strptime(p[0], "%d/%m/%Y") > self.df.index.max()
                ]

            self.points.append(
                (self.df.index.max(), self.df.loc[self.df.index.max(), "wartosc"])
            )

            # interpolation
            dates, values = zip(*self.points)

            # Convert dates to numerical values representing time or elapsed time
            timestamps = np.array(
                [(date - self.max_wibor_real).total_seconds() for date in dates]
            )

            # Create an interpolation function using scipy.interpolate.interp1d
            self.interpolation_function = interp1d(timestamps, values)

            # Convert the new date to a numerical value
            end_timestamp = (self.data_koniec - self.max_wibor_real).total_seconds()

            # Use the interpolation function to estimate the value at the new date
            wibor_value_koniec = self.interpolation_function(end_timestamp).item()

            self.points = [p for p in self.points if p[0] < self.data_koniec]

            self.points.append((self.data_koniec, wibor_value_koniec))

            new_df = pd.DataFrame(self.points, columns=["data", "wartosc"])
            new_df["real"] = "N"
            # new_df['Date'] = pd.to_datetime(new_df['Date'])
            new_df.set_index("data", inplace=True)

            # Concatenate the original DataFrame and the new DataFrame
            dff = pd.concat([self.df, new_df])

        else:
            dff = self.df[self.df.index <= self.data_koniec]

        dff = dff.sort_index()

        self.json_data = [
            {
                "date": dt.datetime.strftime(index, "%Y-%m-%d"),
                "value": value["wartosc"],
                "real": value["real"],
            }
            for index, value in dff.iterrows()
        ]

    def getWibor(self, data) -> float:
        if data > self.max_wibor_real:
            # Generate a new date for interpolation

            # Convert the new date to a numerical value
            new_timestamp = (data - self.max_wibor_real).total_seconds()

            # Use the interpolation function to estimate the value at the new date
            wibor_value = self.interpolation_function(new_timestamp)

        else:
            wibor_value = self._getWiborLastAvailable(data)

        return wibor_value

    def _getWiborLastAvailable(self, data: str) -> float:
        try:
            wibor_value = self.df.loc[data.strftime("%Y-%m-%d")][0]
        except BaseException:
            try:
                wibor_value = self.df[self.df.index < data.strftime("%Y-%m-%d")].iloc[
                    -1
                ][0]
            except BaseException:
                raise Exception("wibor not available")

        return wibor_value

    def isWiborReal(self, data):
        if data >= self.max_wibor_real:
            return "N"
        return "Y"

        return wibor_value

    @property
    def okres(self):
        return self._okres


class Wibor:

    def __init__(self, rodzajWiboru: str, db):

        file_name = ""
        if rodzajWiboru == "3M":
            self._okres = 3
            wib_rodzaj = "wibor3m"
        if rodzajWiboru == "1M":
            self._okres = 1
            wib_rodzaj = "wibor1m"
        elif rodzajWiboru == "6M":
            self._okres = 6
            wib_rodzaj = "wibor6m"
        elif rodzajWiboru == "stopa_ref":
            self._okres = 1
            wib_rodzaj = "stopa_ref"

        self.df = pd.read_sql(
            sql_text(f"SELECT data, wartosc FROM wibor WHERE rodzaj='{wib_rodzaj}'"),
            con=db.engine.connect(),
            index_col="data",
        )

        self.df.index = pd.to_datetime(self.df.index, format="%Y-%m-%d")

        self.df.sort_index(inplace=True)

    def getWibor(self, data: str) -> float:
        return self._getWiborLastAvailable(data)

    def getNearestWibor(self, data: str):
        zmr_iloc_idx = self.df.index.get_indexer([data], method="nearest")
        wibor_zamr_value = self.df.iloc[zmr_iloc_idx].iloc[0][0]
        return wibor_zamr_value

    def getWiborExact(self, data: str):
        try:
            wibor_zamr_value = self.df.loc[data.strftime("%Y-%m-%d")][0]
        except BaseException:
            wibor_zamr_value = None
        return wibor_zamr_value

    def _getWiborLastAvailable(self, data: str) -> float:
        try:
            wibor_value = self.df.loc[data.strftime("%Y-%m-%d")][0]
        except BaseException:
            try:
                wibor_value = self.df[self.df.index < data.strftime("%Y-%m-%d")].iloc[
                    -1
                ][0]
            except BaseException:
                raise Exception("wibor not available")

        return wibor_value

    @property
    def okres(self):
        return self._okres

    @property
    def max_wibor_date(self):
        return self.df.index.max()


def is_business_day(date):
    return bool(len(pd.bdate_range(date, date)))


def next_business_day(date):
    i = 0
    while not is_business_day(date):
        date = date + dt.timedelta(days=1)
        if i > 100:
            raise Exception("next_business_day not found")
    return date


def generateFromWiborFileInter(
    wibor,
    kapital,
    okresy,
    start_date,
    marza,
    transze,
    nadplaty,
    wakacje,
    dni_zmiany_splaty,
    ubezpieczenie_pomostowe_do,
    ubezpieczenie_pomostowe_stopa,
    wibor_value_start,
    tylko_marza=False,
):

    # daty_splaty = [(start_date + relativedelta(months=i)).strftime('%Y-%m-%d') for i in range(okresy+1)]
    daty_splaty = []
    # wakacje= ['2022-08', '2022-09','2022-10','2022-11', '2023-02','2023-05','2023-08','2023-11']

    # nasze_daty_splaty = [(dt.datetime.strptime('2021-11-18', '%Y-%m-%d') + relativedelta(months=i)).strftime('%Y-%m-%d') for i in range(6)] + \
    # [(dt.datetime.strptime('2022-05-04', '%Y-%m-%d') + relativedelta(months=i)).strftime('%Y-%m-%d')  for i in range(3)] + \
    # [(dt.datetime.strptime('2022-08-29', '%Y-%m-%d') + relativedelta(months=i)).strftime('%Y-%m-%d')  for i in range(360)]

    # print(f"nasze daty splaty {nasze_daty_splaty}")

    grosze = decimal.Decimal(".01")

    # 1. dane zwyklego oprocentowania
    # 2. dane o splatach rat
    # 3. generator
    pomost_done = False
    old_wibor_value = 0
    old_wibor_day_business_day = start_date
    wibor_value = 0
    wibor_day_business_day = start_date
    opr_arr = []
    opr_wib = []
    if not tylko_marza:
        for i in range(0, int(okresy / wibor.okres) + 1):
            wibor_day = start_date + relativedelta(months=3 * i)

            old_wibor_day_business_day = wibor_day_business_day
            wibor_day_business_day = wibor_day - BDay(2)

            old_wibor_value = wibor_value

            # check if wibor_value_start is not None

            wibor_value = wibor.getWibor(wibor_day_business_day)

            if wibor_value_start is not None and i == 0:
                wibor_value = wibor_value_start

            if ubezpieczenie_pomostowe_do and not pomost_done:
                if wibor_day < dt.datetime.strptime(
                    ubezpieczenie_pomostowe_do, "%Y-%m-%d"
                ):
                    wibor_value += ubezpieczenie_pomostowe_stopa
                else:
                    opr_wib.append(
                        {
                            "dzien": ubezpieczenie_pomostowe_do,
                            "proc": float(
                                decimal.Decimal(marza + old_wibor_value).quantize(
                                    grosze
                                )
                            ),
                            "real": wibor.isWiborReal(old_wibor_day_business_day),
                        }
                    )
                    opr_arr.append(
                        {
                            "dzien": ubezpieczenie_pomostowe_do,
                            "proc": float(
                                decimal.Decimal(marza + old_wibor_value).quantize(
                                    grosze
                                )
                            ),
                            "real": wibor.isWiborReal(old_wibor_day_business_day),
                            "rodzaj": "wibor",
                            "typ": "",
                        }
                    )
                    pomost_done = True

            # print(f"day: {wibor_day}, value: {wibor_value}")
            opr_wib.append(
                {
                    "dzien": wibor_day.strftime("%Y-%m-%d"),
                    "proc": float(
                        decimal.Decimal(marza + wibor_value).quantize(grosze)
                    ),
                    "real": wibor.isWiborReal(wibor_day_business_day),
                }
            )
            opr_arr.append(
                {
                    "dzien": wibor_day.strftime("%Y-%m-%d"),
                    "proc": float(
                        decimal.Decimal(marza + wibor_value).quantize(grosze)
                    ),
                    "real": wibor.isWiborReal(wibor_day_business_day),
                    "rodzaj": "wibor",
                    "typ": "",
                }
            )

    n = 1
    N = 0
    wakacje_in_progress = False
    miesiac_start_date = start_date.strftime("%Y-%m")
    aktualny_dzien_splaty = start_date
    aktualny_dzien_splaty_dzien = dt.datetime.strftime(start_date, "%d")
    for s in dni_zmiany_splaty:
        if (
            miesiac_start_date == s[0:7]
            and dt.datetime.strptime(s, "%Y-%m-%d") > start_date
        ):
            aktualny_dzien_splaty = dt.datetime.strptime(s, "%Y-%m-%d")
            aktualny_dzien_splaty_dzien = dt.datetime.strftime(
                aktualny_dzien_splaty, "%d"
            )
            break

    while N != okresy:
        # potrzebny nam miesiac
        dzien_splaty = aktualny_dzien_splaty + relativedelta(months=1)

        # month of dzien_splaty
        miesiac = dt.datetime.strftime(dzien_splaty, "%m")
        # rok of dzien_splaty
        rok = dt.datetime.strftime(dzien_splaty, "%Y")

        # sprobujmy utworzyc nowa date splaty
        try:
            str_dzien = f"{rok}-{miesiac}-{aktualny_dzien_splaty_dzien}"
            # print(f"str dzien: {str_dzien}")
            dzien_splaty = dt.datetime.strptime(str_dzien, "%Y-%m-%d")
        except BaseException:
            # zostajemy przy aktualnej dacie splaty
            pass

        miesiac_dnia_splaty = dzien_splaty.strftime("%Y-%m")
        for s in dni_zmiany_splaty:
            if miesiac_dnia_splaty == s[0:7]:
                dzien_splaty = dt.datetime.strptime(s, "%Y-%m-%d")
                aktualny_dzien_splaty_dzien = dt.datetime.strftime(dzien_splaty, "%d")
                break

        dzien_splaty_business_day = dzien_splaty
        # check if dzien_splaty is business day
        if not is_business_day(dzien_splaty_business_day):
            dzien_splaty_business_day = next_business_day(dzien_splaty_business_day)

        ########################################
        # WAKACJE KREYDTOWE
        ########################################

        # dzien_splaty = dt.datetime.strptime(nasze_daty_splaty[n], '%Y-%m-%d')
        # check if dzien_splaty is not same month and day as wakacje
        if not (dzien_splaty.strftime("%Y-%m") in wakacje):
            N += 1

            daty_splaty.append(dzien_splaty_business_day.strftime("%Y-%m-%d"))
            if wakacje_in_progress:
                wakacje_in_progress = False
                # opr_wib.append({"dzien":dzien_splaty.strftime('%Y-%m-%d'), "proc": float(decimal.Decimal(marza+wibor.getWibor(dzien_splaty)).quantize(grosze))})
                opr_arr.append(
                    {
                        "dzien": dzien_splaty_business_day.strftime("%Y-%m-%d"),
                        "proc": float(
                            decimal.Decimal(
                                marza + wibor.getWibor(dzien_splaty_business_day)
                            ).quantize(grosze)
                        ),
                        "real": wibor.isWiborReal(dzien_splaty_business_day),
                        "rodzaj": "splata",
                        "typ": "",
                    }
                )
        else:
            wakacje_in_progress = True
            # opr_wib.append({"dzien":dzien_splaty.strftime('%Y-%m-%d'), "proc": float(decimal.Decimal(0).quantize(grosze))})
            opr_arr.append(
                {
                    "dzien": dzien_splaty_business_day.strftime("%Y-%m-%d"),
                    "proc": float(
                        decimal.Decimal(
                            marza + wibor.getWibor(dzien_splaty_business_day)
                        ).quantize(grosze)
                    ),
                    "real": wibor.isWiborReal(dzien_splaty_business_day),
                    "rodzaj": "splata",
                    "typ": "W",
                }
            )
        aktualny_dzien_splaty = dzien_splaty
        n += 1

    # sort opr_arr by dzien
    opr_arr = sorted(opr_arr, key=lambda i: i["dzien"])

    # data frame from opr_wib
    df = pd.DataFrame(opr_wib)
    df["dzien"] = pd.to_datetime(df["dzien"])

    # df sorted by dzien
    df = df.sort_values(by=["dzien"])

    # print(df)

    # oprocentowanie z uwzglednieniem wakacji kredytowych
    opr_wakacje = []
    wakacje = False
    for r in opr_arr:
        if r["rodzaj"] == "splata" and r["typ"] == "W":
            opr_wakacje.append(
                {
                    "dzien": r["dzien"],
                    "proc": float(decimal.Decimal(0).quantize(grosze)),
                    "real": r["real"],
                }
            )
            wakacje = True
        else:
            if wakacje:
                if r["rodzaj"] == "splata" and r["typ"] == "":
                    wibor_value = df[df["dzien"] < r["dzien"]]["proc"].iloc[-1]

                    opr_wakacje.append(
                        {"dzien": r["dzien"], "proc": wibor_value, "real": r["real"]}
                    )
                    wakacje = False

            else:
                opr_wakacje.append(
                    {"dzien": r["dzien"], "proc": r["proc"], "real": r["real"]}
                )

    # print(opr_wakacje)

    transze_out = []
    if transze:
        for tr in transze:
            transze_out.append(
                {"dzien": tr["dzien"].strftime("%Y-%m-%d"), "kapital": tr["wartosc"]}
            )

    if tylko_marza:
        p_start = float(decimal.Decimal(marza).quantize(grosze))
    else:
        p_start = float(
            decimal.Decimal(wibor.getWibor(start_date) + marza).quantize(grosze)
        )

    data = {
        "K": kapital,
        "transze": transze_out,
        "nadplaty": nadplaty,
        "p": p_start,
        "marza": marza,
        "start": start_date.strftime("%Y-%m-%d"),
        "daty_splaty": daty_splaty,
        "oprocentowanie": opr_wakacje,
    }

    return data



"""
1. krzywa typu splajn posiada punkt startowy, końcowy, oraz punkty pośrednie
2. do każdego punktu przypisany jest timestamp oraz wartość
3. krzywą można interpolować
"""

from dataclasses import dataclass
from datetime import datetime
from typing import List, Tuple

import numpy as np
from dateutil.relativedelta import relativedelta
from scipy.interpolate import CubicSpline, interp1d

from procenty.utils import diff_month


@dataclass
class Krzywa:

    punkty: List[Tuple[datetime, float]]

    def __post_init__(self):

        self.punkty.sort(key=lambda x: x[0])

        self.start = self.punkty[0][0]
        self.end = self.punkty[-1][0]

        self.timestamps = np.array([dt.timestamp() for dt, _ in self.punkty])
        self.values = np.array([val for _, val in self.punkty])

        spline = CubicSpline(self.timestamps, self.values)

        # Generate points for smooth plot
        self.timestamps_smooth = np.linspace(
            self.timestamps.min(), self.timestamps.max(), 500
        )
        self.values_smooth = spline(self.timestamps_smooth)

        self.interpolation_function = interp1d(
            self.timestamps_smooth, self.values_smooth
        )

    def __repr__(self):
        return f"Krzywa({self.start}, {self.end}, {self.punkty})"

    def __mul__(self, inflator):
        from procenty.inflacja import Inflacja

        if isinstance(inflator, Inflacja):

            wynik = []
            punkty_miesiac = self.podzial_miesiac()

            for punkt in punkty_miesiac:
                wartosc_inflator = inflator.urealnij(punkt[0], punkt[1])
                wynik.append((punkt[0], wartosc_inflator))

            return wynik
        return NotImplemented

    @property
    def splajn(self):
        return list(zip(self.timestamps_smooth, self.values_smooth))

    @property
    def punkty_zip(self):
        return list(zip(self.timestamps, self.values))

    def podzial(self, okres_dni: int) -> list:
        """
        Dzieli krzywą na okresy o długości okres_dni
        """

        dajs = [
            self.start + relativedelta(days=(i + 1) * okres_dni)
            for i in range(int((self.end - self.start).days / okres_dni))
        ]
        opr = [(d, self.interpolation_function(d.timestamp()).item()) for d in dajs]

        return opr

    def podzial_miesiac(self) -> list:
        """
        Na razie dzieli na miesace biorac konkretny dzień.
        Trzeba zrobić uśrednienie z całego miesiąca.
        """
        dajs = [self.start]
        dajs.extend(
            [
                self.start + relativedelta(months=(i + 1))
                for i in range(diff_month(self.end, self.start))
            ]
        )
        opr = [(d, self.interpolation_function(d.timestamp()).item()) for d in dajs]

        return opr
